
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>

#include <iostream>
#include <windows.h>
#include <conio.h>

// Thread block size
const int BLOCK_SIDE = 16;		//  -
const int MATRIX_SIDE = 4096;

const int BLOCK_SIZE = BLOCK_SIDE * BLOCK_SIDE;
const int MATRIX_SIZE = MATRIX_SIDE * MATRIX_SIDE;



__global__ void transposeKernel_fast(const float* matrix, float* matrixT, int width, int height)
{
	__shared__ float tile[BLOCK_SIDE][BLOCK_SIDE + 1];

	int xIndex = blockIdx.x * BLOCK_SIDE + threadIdx.x; 
	int yIndex = blockIdx.y * BLOCK_SIDE + threadIdx.y; 
	int index_in = xIndex + (yIndex) * width;

	xIndex = blockIdx.y * BLOCK_SIDE + threadIdx.x; 
	yIndex = blockIdx.x * BLOCK_SIDE + threadIdx.y; 
	int index_out = xIndex + (yIndex) * height;

	tile[threadIdx.y][threadIdx.x] = matrix[index_in];

	__syncthreads();

	matrixT[index_out] = tile[threadIdx.x][threadIdx.y];
}


__global__ void transposeKernel_simple(const float* matrix, float* matrixT, int width, int height)
{
	int xIndex = blockIdx.x * BLOCK_SIDE + threadIdx.x; 
	int yIndex = blockIdx.y * BLOCK_SIDE + threadIdx.y; 

	int index_in = xIndex + (yIndex) * width;
	int index_out = yIndex + (xIndex) * width;

	matrixT[index_out] = matrix[index_in];
}

unsigned __int64 transposeWithCuda_fast(const float* matrix, float* matrixT, int width, int height);
unsigned __int64 transposeWithCuda_simple(const float* matrix, float* matrixT, int width, int height);
void printfMatrix(const float* matrix, int width, int height);
void transposeCPU(const float* matrix, float* matrixT, int width, int height);
bool isSame (float* array1, float* array2, int size);


int main()
{
    int width = MATRIX_SIDE;
	int height = MATRIX_SIDE;
	int matrixSize = width * height;
	
	float* matrix_original = (float*)malloc(matrixSize * sizeof(float));
	float* matrixT_CPU = (float*)malloc(matrixSize * sizeof(float));
	float* matrixT_GPU = (float*)malloc(matrixSize * sizeof(float));

	if(matrix_original == NULL || matrixT_CPU == NULL || matrixT_GPU == NULL)
	{
		fprintf(stderr, "Malloc error");
		exit(EXIT_FAILURE);
	}

	//  
	for(int i = 0; i < matrixSize; i++)	
		matrix_original[i] = (float) i;

	//   CPU
	unsigned __int64 numberOfCycles = __rdtsc();
	transposeCPU(matrix_original, matrixT_CPU, width, height);
	numberOfCycles = __rdtsc() - numberOfCycles;

	printf("\nCPU transpose:\n\t\t%ld\n", numberOfCycles);


	//   GPU
    numberOfCycles = transposeWithCuda_simple(matrix_original, matrixT_GPU, width, height);
    printf("\nCUDA simple transpose:\n\t\t%ld", numberOfCycles);

	if(isSame(matrixT_CPU, matrixT_GPU, matrixSize))
		printf("\nResult is the same as on CPU.\n");
	else printf("\nresult is NOT the same as on CPU.\n");
	
	//   GPU
	numberOfCycles = transposeWithCuda_fast(matrix_original, matrixT_GPU, width, height);
	printf("\nCUDA optimized transpose:\n\t\t%ld", numberOfCycles);

	if(isSame(matrixT_CPU, matrixT_GPU, matrixSize))
		printf("\nResult is the same as on CPU.\n");
	else printf("\nresult is NOT the same as on CPU.\n");


	free(matrix_original);
	free(matrixT_CPU);
	free(matrixT_GPU);

	cudaError_t cudaStatus = cudaDeviceReset();
	if (cudaStatus != cudaSuccess) 
	{
		fprintf(stderr, "cudaDeviceReset failed!");
		return 1;
	}
	_getch();
    return 0;
}


void transposeCPU(const float* matrix, float* matrixT, int width, int height)
{
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
			matrixT[i * width + j] = matrix[j * width + i];
	}
}


unsigned __int64 transposeWithCuda_fast(const float* matrix_original, float* matrix_transposed, int width, int height)
{
    cudaError_t cudaStatus;
	
	size_t sizeInBytes = width * height * sizeof(float);

	float* d_matrix_original;
	cudaStatus = cudaMalloc((void**)&d_matrix_original, sizeInBytes);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMalloc failed!");
		exit(EXIT_FAILURE);
	}

	cudaStatus = cudaMemcpy(d_matrix_original, matrix_original, sizeInBytes, cudaMemcpyHostToDevice);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMemcpy failed!");
		exit(EXIT_FAILURE);
	}

	float* d_matrix_transposed;
	cudaStatus = cudaMalloc((void**)&d_matrix_transposed, sizeInBytes);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMalloc failed!");
		exit(EXIT_FAILURE);
	}

	dim3 dimBlock(BLOCK_SIDE, BLOCK_SIDE);					//  
	dim3 dimGrid(width / BLOCK_SIDE, height / BLOCK_SIDE);	//   
    
	unsigned __int64 numberOfCycles = __rdtsc();

	transposeKernel_fast<<<dimGrid, dimBlock>>>(d_matrix_original, d_matrix_transposed, width, height);
	cudaDeviceSynchronize();

	numberOfCycles = __rdtsc() - numberOfCycles;

    // Copy output vector from GPU buffer to host memory.
    cudaStatus = cudaMemcpy(matrix_transposed, d_matrix_transposed, sizeInBytes, cudaMemcpyDeviceToHost);
    if (cudaStatus != cudaSuccess) {
        fprintf(stderr, "cudaMemcpy failed!");
		exit(EXIT_FAILURE);
    }

    cudaFree(d_matrix_original);
    cudaFree(d_matrix_transposed);
    
    return numberOfCycles;
}

unsigned __int64 transposeWithCuda_simple(const float* matrix_original, float* matrix_transposed, int width, int height)
{
	cudaError_t cudaStatus;

	size_t sizeInBytes = width * height * sizeof(float);

	float* d_matrix_original;
	cudaStatus = cudaMalloc((void**)&d_matrix_original, sizeInBytes);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMalloc failed!");
		exit(EXIT_FAILURE);
	}

	cudaStatus = cudaMemcpy(d_matrix_original, matrix_original, sizeInBytes, cudaMemcpyHostToDevice);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMemcpy failed!");
		exit(EXIT_FAILURE);
	}

	float* d_matrix_transposed;
	cudaStatus = cudaMalloc((void**)&d_matrix_transposed, sizeInBytes);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMalloc failed!");
		exit(EXIT_FAILURE);
	}

	dim3 dimBlock(BLOCK_SIDE, BLOCK_SIDE);					//  
	dim3 dimGrid(width / BLOCK_SIDE, height / BLOCK_SIDE);	//   

	unsigned __int64 numberOfCycles = __rdtsc();

	transposeKernel_simple<<<dimGrid, dimBlock>>>(d_matrix_original, d_matrix_transposed, width, height);
	cudaDeviceSynchronize();

	numberOfCycles = __rdtsc() - numberOfCycles;

	// Copy output vector from GPU buffer to host memory.
	cudaStatus = cudaMemcpy(matrix_transposed, d_matrix_transposed, sizeInBytes, cudaMemcpyDeviceToHost);
	if (cudaStatus != cudaSuccess) {
		fprintf(stderr, "cudaMemcpy failed!");
		exit(EXIT_FAILURE);
	}

	cudaFree(d_matrix_original);
	cudaFree(d_matrix_transposed);

	return numberOfCycles;
}


void printfMatrix(const float* matrix, int width, int height)
{
	for(int i = 0; i < height; i++)
	{
		for (int j = 0; j < width; j++)		
			printf("%.1f\t", matrix[i * width + j]);

		printf("\n");
	}
}


bool isSame (float* array1, float* array2, int size)
{
	bool same = true;
	for(int i = 0; i < size; i++)
	{
		if(array1[i] != array2[i])
		{
			same = false;
			break;
		}
	}
	return same;
}
