#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <windows.h>
#include <math.h>
#include <stdlib.h>
#include <stdint.h>
#include "Shlwapi.h"
#include "hr_time.h"
#include "bmp.h"

#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, char *file, int line, bool abort=true)
{
	if (code != cudaSuccess) 
	{
		fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
		if (abort) exit(code);
	}
}


#define BLOCK_SIZE_X 64
#define BLOCK_SIZE_Y 8

BYTE* DiscreteLaplaceOperator_CPU(BYTE* pixelData, int width, int byteWidth, int height);
BYTE* DiscreteLaplaceOperator_CUDA(BYTE* pixelData_RGB, int width, int byteWidth, int height, CStopWatch& timer);
bool isArrayEqual(BYTE* array1, BYTE* array2, int size);

//     
const int laplaceConvolutionKernel[3][3] = 
{
	{1,  1,  1},
	{1, -8,  1},
	{1,  1,  1}
};


__device__ __constant__ int laplaceConvolutionKernel_gpu[3][3] = 
{
	{1,  1,  1},
	{1, -8,  1},
	{1,  1,  1}
};


__global__ void LaplaceKernel(BYTE* inputBitmap, BYTE* outputBitmap, int trueByteWidth, int height, int dwordWidth, int byteWidth)
{
	const int xIndex = blockIdx.x * BLOCK_SIZE_X + threadIdx.x; 
	const int yIndex = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y; 
	const int index = xIndex + yIndex * dwordWidth;

	if(xIndex >= dwordWidth || yIndex >= height)
		return;

	__shared__ uint32_t tile[BLOCK_SIZE_Y][BLOCK_SIZE_X];

	tile[threadIdx.y][threadIdx.x] = ((uint32_t*)inputBitmap)[index];

	__syncthreads();

	int threadPosX = threadIdx.x * 4;
	int threadAbsPosX = xIndex * 4;

	uint32_t result = 0;

	for (int k = 0; k < 4; k++)
	{
		int bytePosX = threadPosX + k;
		int byteAbsPosX = threadAbsPosX + k;

		if(bytePosX > trueByteWidth)
			break;

		int posX, posY, offsetX, offsetY, absPosX, absPosY;
		int sum = 0;

		for (int j = 0; j < 3; j++)
		{
			for (int i = 0; i < 3; i++)
			{
				offsetX = (i - 1) * 3;
				offsetY = (j - 1);

				posX = bytePosX + offsetX;
				posY = threadIdx.y + offsetY;

				absPosX = byteAbsPosX + offsetX;
				absPosY = yIndex + offsetY;

				//     
				if(posX >= 0 && posX < BLOCK_SIZE_X * 4 && absPosX < trueByteWidth
					&& posY >= 0 && posY < BLOCK_SIZE_Y && absPosY < height)	
				{
					sum += ((byte*)tile[posY])[posX] * laplaceConvolutionKernel_gpu[j][i];
				}
				//    ,      (x / y),    
				else if((absPosX < 0 || absPosX >= trueByteWidth || absPosY < 0 || absPosY >= height) 
					&& ((posX >= 0 && posX < BLOCK_SIZE_X * 4 && absPosX < trueByteWidth) 
					|| (posY >= 0 && posY < BLOCK_SIZE_Y && absPosY < height)))
				{
					if(absPosX < 0 || absPosX >= trueByteWidth) 
						posX = bytePosX;

					if(absPosY < 0 || absPosY >= height) 
						posY = threadIdx.y;	

					sum += ((byte*)tile[posY])[posX] * laplaceConvolutionKernel_gpu[j][i];
				}
				//     ,   ,    
				else
				{
					if(absPosX < 0 || absPosX >= trueByteWidth) 
						absPosX = byteAbsPosX;

					if(absPosY < 0 || absPosY >= height) 
						absPosY = yIndex;	

					sum += inputBitmap[absPosX + absPosY * byteWidth] * laplaceConvolutionKernel_gpu[j][i];	
				}
			}
		}
		if(sum < 0) sum *= -1;
		if(sum > 255) sum = 255;

		((byte*)&result)[k] = sum;
	}

	((uint32_t*)outputBitmap)[index] = result;
}



int main(int argc, char** argv)
{
	if(argc != 3)
	{
		fprintf(stderr, "Usage: %s <inputBMP24file> <outputBMP24file>\n", PathFindFileName(argv[0]));
		return EXIT_FAILURE;
	}

	char* fileNameToOpen = argv[1];
	char* fileNameToSave = argv[2];

	int imageWidth, imageHeight;
	long paddedSize;
	BYTE* pixelData = LoadBMP(&imageWidth, &imageHeight, &paddedSize, fileNameToOpen);
	if(pixelData == NULL)
	{
		fprintf(stderr, "BMP load error.\n");
		return EXIT_FAILURE;
	}
	int byteWidth = imageWidth * 3;
	if (byteWidth % 4 != 0)
		byteWidth = 4 * (byteWidth / 4 + 1);

	CStopWatch timer1, timer2;

	timer1.startTimer();
	BYTE* newImageCPU = DiscreteLaplaceOperator_CPU(pixelData, imageWidth, byteWidth, imageHeight);
	timer1.stopTimer();

	if(newImageCPU == NULL)
	{	
		delete[] pixelData;
		fprintf(stderr, "DiscreteLaplaceOperator_CPU() error.\n");
		return EXIT_FAILURE;
	}
	printf("CPU: %.0f milliseconds\n", floor(timer1.getElapsedTime_milliseconds() + 0.5));

	BYTE* newImageCUDA = DiscreteLaplaceOperator_CUDA(pixelData, imageWidth, byteWidth, imageHeight, timer2);
	if(newImageCUDA == NULL)
	{	
		delete[] pixelData;
		delete[] newImageCPU;
		fprintf(stderr, "DiscreteLaplaceOperator_CUDA() error.\n");
		return EXIT_FAILURE;
	}
	printf("CUDA: %.0f milliseconds\n", floor(timer2.getElapsedTime_milliseconds() + 0.5));

	//    
	if(isArrayEqual(newImageCPU, newImageCUDA, paddedSize) == false)
	{
		delete[] pixelData;
		delete[] newImageCPU;
		delete[] newImageCUDA;

		fprintf(stderr, "Results of CPU and CUDA are NOT equal!!!\n");
		return EXIT_FAILURE;
	}
	else printf("Results are equal. Saving to %s...\n", PathFindFileName(fileNameToSave));

	//------------------------SAVING RESULT-------------------------------------

	bool isOk = SaveBMP(newImageCUDA, imageWidth, imageHeight, paddedSize, fileNameToSave);
	delete[] pixelData;
	delete[] newImageCPU;
	delete[] newImageCUDA;

	if(isOk == false)
	{
		fprintf(stderr, "Save BMP error.\n");
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}


//     CPU
BYTE* DiscreteLaplaceOperator_CPU(BYTE* pixelData, int width, int byteWidth, int height)
{
	BYTE* result = new BYTE[byteWidth * height]();
	if(result == NULL)
		return NULL;

	int pos;
	for (int y = 0; y < height; y++) //  
	{
		for (int x = 0; x < width; x++) //   
		{
			int rSum = 0, gSum = 0, bSum = 0;

			for (int j = 0; j < 3; j++)
			{
				for (int i = 0; i < 3; i++)
				{
					int pixelPosX = x + (i - 1);
					int pixelPosY = y + (j - 1);

					if (pixelPosX == width) pixelPosX = width - 1;
					if (pixelPosX == -1) pixelPosX = 0;
					if (pixelPosY == height) pixelPosY = height - 1;
					if (pixelPosY == -1) pixelPosY = 0;

					pos = byteWidth * pixelPosY + pixelPosX * 3;
					int kernelVal = laplaceConvolutionKernel[j][i];

					rSum += pixelData[pos + 0] * kernelVal;
					gSum += pixelData[pos + 1] * kernelVal;
					bSum += pixelData[pos + 2] * kernelVal;

					//kSum += kernelVal;
				}
			}
			if (rSum < 0) rSum *= -1;
			if (rSum > 255) rSum = 255;

			if (gSum < 0) gSum *= -1;
			if (gSum > 255) gSum = 255;

			if (bSum < 0) bSum *= -1;
			if (bSum > 255) bSum = 255;

			//    
			pos = byteWidth * y + x * 3;
			result[pos + 0] = (byte)rSum;
			result[pos + 1] = (byte)gSum;
			result[pos + 2] = (byte)bSum;
		}
	}
	return result;
}


//     CUDA GPU
BYTE* DiscreteLaplaceOperator_CUDA(BYTE* pixelData_RGB, int width, int byteWidth, int height, CStopWatch& timer)
{
	size_t size = byteWidth * height;

	BYTE *pixelData_gpu, *pixelDataResult_gpu;

	gpuErrchk( cudaMalloc((void**)&pixelData_gpu, size) );
	gpuErrchk( cudaMalloc((void**)&pixelDataResult_gpu, size) );

	gpuErrchk( cudaMemcpy(pixelData_gpu, pixelData_RGB, size, cudaMemcpyHostToDevice) );

	//-----------------------------
	int dwordWidth = byteWidth / 4;
	int trueByteWidth = width * 3;

	int gridSize_X = ceil((double)dwordWidth / (double)BLOCK_SIZE_X);
	int gridSize_Y = ceil((double)height / (double)BLOCK_SIZE_Y);
	dim3 dimGrid(gridSize_X, gridSize_Y);
	dim3 dimBlock(BLOCK_SIZE_X, BLOCK_SIZE_Y);

	timer.startTimer();
	LaplaceKernel<<<dimGrid, dimBlock>>> (pixelData_gpu, pixelDataResult_gpu, trueByteWidth, height, dwordWidth, byteWidth);

	gpuErrchk( cudaPeekAtLastError() );
	gpuErrchk( cudaDeviceSynchronize() );

	timer.stopTimer();
	//----------------------------
	BYTE* result = new BYTE[size];
	if (result == NULL)
	{
		fprintf(stderr, "new[] failed!");
		cudaDeviceReset();
		return NULL;
	}

	gpuErrchk( cudaMemcpy(result, pixelDataResult_gpu, size, cudaMemcpyDeviceToHost) );

	gpuErrchk( cudaFree(pixelData_gpu) );
	gpuErrchk( cudaFree(pixelDataResult_gpu) );

	gpuErrchk( cudaDeviceReset() );
	return result;
}


//  
bool isArrayEqual(BYTE* array1, BYTE* array2, int size)
{
	bool isEqual = true;

	for (int i = 0; i < size; i++)
	{
		if(array1[i] != array2[i])
		{
			isEqual = false;
			break;
		}
	}
	return isEqual;
}