
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#   pragma warning( disable : 4996 ) // disable deprecated warning 
#endif

#pragma pack(1)

typedef struct
{
    short type;
    int size;
    short reserved1;
    short reserved2;
    int offset;
} BMPHeader;

typedef struct
{
    int size;
    int width;
    int height;
    short planes;
    short bitsPerPixel;
    unsigned compression;
    unsigned imageSize;
    int xPelsPerMeter;
    int yPelsPerMeter;
    int clrUsed;
    int clrImportant;
} BMPInfoHeader;



//Isolated definition
typedef struct
{
    unsigned char x, y, z, w;
} uchar4;



extern "C" void LoadBMPFile(uchar4 **dst, int *width, int *height, const char *name)
{
    BMPHeader hdr;
    BMPInfoHeader infoHdr;
    int x, y;

    FILE *fd;


    printf("Loading %s...\n", name);

    if (sizeof(uchar4) != 4)
    {
        printf("***Bad uchar4 size***\n");
        exit(EXIT_SUCCESS);
    }

    if (!(fd = fopen(name,"rb")))
    {
        printf("***BMP load error: file access denied***\n");
        exit(EXIT_SUCCESS);
    }

    fread(&hdr, sizeof(hdr), 1, fd);

    if (hdr.type != 0x4D42)
    {
        printf("***BMP load error: bad file format***\n");
        exit(EXIT_SUCCESS);
    }

    fread(&infoHdr, sizeof(infoHdr), 1, fd);

    if (infoHdr.bitsPerPixel != 24)
    {
        printf("***BMP load error: invalid color depth***\n");
        exit(EXIT_SUCCESS);
    }

    if (infoHdr.compression)
    {
        printf("***BMP load error: compressed image***\n");
        exit(EXIT_SUCCESS);
    }

    *width  = infoHdr.width;
    *height = infoHdr.height;
    *dst    = (uchar4 *)malloc(*width * *height * 4);

    printf("BMP width: %u\n", infoHdr.width);
    printf("BMP height: %u\n", infoHdr.height);

    fseek(fd, hdr.offset - sizeof(hdr) - sizeof(infoHdr), SEEK_CUR);

    for (y = 0; y < infoHdr.height; y++)
    {
        for (x = 0; x < infoHdr.width; x++)
        {
            (*dst)[(y * infoHdr.width + x)].z = fgetc(fd);
            (*dst)[(y * infoHdr.width + x)].y = fgetc(fd);
            (*dst)[(y * infoHdr.width + x)].x = fgetc(fd);
        }

        for (x = 0; x < (4 - (3 * infoHdr.width) % 4) % 4; x++)
            fgetc(fd);
    }


    if (ferror(fd))
    {
        printf("***Unknown BMP load error.***\n");
        free(*dst);
        exit(EXIT_SUCCESS);
    }
    else
        printf("BMP file loaded successfully!\n");

    fclose(fd);
}


extern "C" void SaveBMPFile(uchar4 *src, int width, int height, const char *name)
{
	BMPHeader hdr;
	BMPInfoHeader infoHdr;
	int x, y;

	FILE *fd;

	printf("Saving %s...\n", name);

	if (sizeof(uchar4) != 4)
	{
		printf("***Bad uchar4 size***\n");
		exit(EXIT_SUCCESS);
	}

	if (!(fd = fopen(name,"wb")))
	{
		printf("***BMP load error: file access denied***\n");
		exit(EXIT_SUCCESS);
	}

	// andinitialize them to zero
	memset ( &hdr, 0, sizeof (BMPHeader ) );
	memset ( &infoHdr, 0, sizeof (BMPInfoHeader ) );

	int byteWidth = width * 3;
	if (byteWidth % 4 != 0)
		byteWidth = 4 * (byteWidth / 4 + 1);
	int paddedsize = byteWidth * height;

	// fill the fileheader with data
	hdr.type = 0x4d42;       // 0x4d42 = 'BM'
	hdr.reserved1 = 0;
	hdr.reserved2 = 0;
	hdr.size = sizeof(BMPHeader) + sizeof(BMPInfoHeader) + paddedsize;
	hdr.offset = 0x36;		// number of bytes to start of bitmap bits

	// fill the infoheader

	infoHdr.size = sizeof(BMPInfoHeader);
	infoHdr.width = width;
	infoHdr.height = height;
	infoHdr.planes = 1;			// we only have one bitplane
	infoHdr.bitsPerPixel = 24;		// RGB mode is 24 bits
	infoHdr.compression = 0;	
	infoHdr.imageSize = 0;		// can be 0 for 24 bit images
	infoHdr.xPelsPerMeter = 0x0ec4;     // paint and PSP use this values
	infoHdr.yPelsPerMeter = 0x0ec4;     
	infoHdr.clrUsed = 0;			// we are in RGB mode and have no palette
	infoHdr.clrImportant = 0;    // all colors are important

	fwrite(&hdr, sizeof(hdr), 1, fd);
	fwrite(&infoHdr, sizeof(infoHdr), 1, fd);

	fseek(fd, hdr.offset - sizeof(hdr) - sizeof(infoHdr), SEEK_CUR);

	for (y = 0; y < infoHdr.height; y++)
	{
		for (x = 0; x < infoHdr.width; x++)
		{
			fputc(src[(y * infoHdr.width + x)].z, fd);
			fputc(src[(y * infoHdr.width + x)].y, fd);
			fputc(src[(y * infoHdr.width + x)].x, fd);
		}

		for (x = 0; x < (4 - (3 * infoHdr.width) % 4) % 4; x++)
			fputc(0, fd);
	}


	if (ferror(fd))
	{
		printf("***Unknown BMP save error.***\n");
		exit(EXIT_SUCCESS);
	}
	else
		printf("BMP file saved successfully!\n");

	fclose(fd);
}