#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 

#include "headers.h"



#define LAPLASE_KERNEL_SIZE 3

#define BLOCKDIM_X 64
#define BLOCKDIM_Y 8

__device__ __constant__ int laplaceConvolutionKernel_gpu[LAPLASE_KERNEL_SIZE][LAPLASE_KERNEL_SIZE] = 
{
	{1,  1,  1},
	{1, -8,  1},
	{1,  1,  1}
};

inline __device__ unsigned char clamp(int value){
	unsigned int v;
	asm("cvt.sat.u8.s32 %0, %1;" : "=r"(v) : "r"(value));
	return v;
}


//---------------------COPY KERNEL---------------------------------
__global__ void Copy(TColor *dst,TColor *src, int imageW, int imageH)
{
	const int ix = blockDim.x * blockIdx.x + threadIdx.x;
	const int iy = blockDim.y * blockIdx.y + threadIdx.y;

	if (ix < imageW && iy < imageH)
	{
		dst[imageW * iy + ix] = src[imageW * iy + ix];
	}
}


extern "C" void
	cuda_Copy(TColor *d_dst, TColor *cuda_Src, int imageW, int imageH)
{
	dim3 threads(BLOCKDIM_X, BLOCKDIM_Y);

	int gridSize_X = ceil((double)imageW / (double)BLOCKDIM_X);
	int gridSize_Y = ceil((double)imageH / (double)BLOCKDIM_Y);
	dim3 grid(gridSize_X, gridSize_Y);

	Copy<<<grid, threads>>>(d_dst, cuda_Src, imageW, imageH);
}
/**********************************************************************************************************/



//---------------------LAPLASE KERNEL---------------------------------
__global__ void Laplace(TColor *dst, TColor *src, int imageW, int imageH)
{
	__shared__ TColor tile[BLOCKDIM_Y][BLOCKDIM_X];

	const int ix = (blockDim.x - 2) * blockIdx.x + threadIdx.x;
	const int iy = (blockDim.y - 2) * blockIdx.y + threadIdx.y;
	TColor tmp = 0;

	if (ix < imageW && iy < imageH)
	{		
		tile[threadIdx.y][threadIdx.x] = src[imageW * iy + ix];
		__syncthreads();

		if( (ix == 0 && iy == 0) || (ix == 0 && threadIdx.x == 0 && threadIdx.y >= 1 && threadIdx.y <= BLOCKDIM_Y - 2) 
			|| (iy == 0 && threadIdx.y == 0 && threadIdx.x >= 1 && threadIdx.x <= BLOCKDIM_X - 2)
			|| (threadIdx.x >=1 && threadIdx.x <= BLOCKDIM_X - 2 && threadIdx.y >=1 && threadIdx.y <= BLOCKDIM_Y - 2))
		{
			//Result accumulator
			int r = 0, g = 0, b = 0;
			for(int i = 0; i < LAPLASE_KERNEL_SIZE; i++)
			{
				for(int j = 0; j < LAPLASE_KERNEL_SIZE; j++)
				{
					int posX = threadIdx.x + i - 1;
					int posY = threadIdx.y + j - 1;

					int absPosX = ix + i - 1;
					int absPosY = iy + j - 1;

					posX = (absPosX < 0 || absPosX >= imageW) ? threadIdx.x : posX;
					posY = (absPosY < 0 || absPosY >= imageH) ? threadIdx.y : posY;

					r += ((unsigned char*)&tile[posY][posX])[0] * laplaceConvolutionKernel_gpu[j][i];
					g += ((unsigned char*)&tile[posY][posX])[1] * laplaceConvolutionKernel_gpu[j][i];
					b += ((unsigned char*)&tile[posY][posX])[2] * laplaceConvolutionKernel_gpu[j][i];
				}
			}
			//
			// val < 0 -> 0
			// val > 255 -> 255
			//r = clamp(r);
			//g = clamp(g);
			//b = clamp(b);
			r = (r < 0) ? r*(-1) : r;
			r = (r > 255) ? 255 : r;
			g = (g < 0) ? g*(-1) : g;
			g = (g > 255) ? 255 : g;
			b = (b < 0) ? b*(-1) : b;
			b = (b > 255) ? 255 : b;

			((unsigned char*)&tmp)[0] = r;
			((unsigned char*)&tmp)[1] = g;
			((unsigned char*)&tmp)[2] = b;
			((unsigned char*)&tmp)[3] = 0;

			dst[imageW * iy + ix] = tmp;
		}

	}
}

extern "C"
	double cuda_Laplace(TColor *d_dst, TColor *cuda_Src, int imageW, int imageH)
{
	dim3 threads(BLOCKDIM_X, BLOCKDIM_Y);

	int gridSize_X = ceil((double)imageW / (double)(BLOCKDIM_X - 2)) + 1;
	int gridSize_Y = ceil((double)imageH / (double)(BLOCKDIM_Y - 2)) + 1;
	dim3 grid(gridSize_X, gridSize_Y);

	CStopWatch timer;
	timer.startTimer();

	Laplace<<<grid, threads>>>(d_dst, cuda_Src, imageW, imageH);

	checkCudaErrors ( cudaDeviceSynchronize() );
	timer.stopTimer();

	return timer.getElapsedTime_milliseconds();
}
