
// OpenGL Graphics includes
#include <GL/glew.h>
#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/freeglut.h>
#endif

// CUDA utilities and system includes
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>

// Includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "headers.h"
#include "DiscreteLaplaceOperator_CPU.h"



////////////////////////////////////////////////////////////////////////////////
// Global data handlers and parameters
////////////////////////////////////////////////////////////////////////////////
//OpenGL PBO and texture "names"
GLuint gl_PBO, gl_Tex;
struct cudaGraphicsResource *cuda_pbo_resource; // handles OpenGL-CUDA exchange
//Source image on the host side
_uchar4 *h_Src;
int imageW, imageH;

//Source image on gpu
TColor *cuda_Src;

////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////
int  g_Kernel = 0;

bool isModeChanged = false;

#define BUFFER_DATA(i) ((char *)0 + i)


//  
template<typename T>
bool isArrayEqual(T* array1, T* array2, int size)
{
	bool isEqual = true;

	for (int i = 0; i < size; i++)
	{
		if(!(array1[i] == array2[i]))
		{
			isEqual = false;
			break;
		}
	}
	return isEqual;
}


void runImageFilters(TColor *d_dst, TColor *cuda_Src)
{
	//printf("Filters runned!!!\n");
	double time;
	switch (g_Kernel)
	{
	case 0:
		cuda_Copy(d_dst, cuda_Src, imageW, imageH);
		break;

	case 1:
		time = cuda_Laplace(d_dst, cuda_Src, imageW, imageH);
		printf("Cuda time: %f\n", time);
		break;

	case 2:
		{
			_uchar4* cpuResult = (_uchar4*)malloc(imageH * imageW * 4);
			time = cpu_Laplace_withTime(cpuResult, h_Src, imageW, imageH);
			checkCudaErrors( cudaMemcpy(d_dst, cpuResult, imageW * imageH * 4, cudaMemcpyHostToDevice) );
			free(cpuResult);

			printf("CPU time: %f\n", time);
		}		
		break;

	// Compare
	case 3:
		{
			_uchar4* cpuResult = (_uchar4*)malloc(imageH * imageW * 4);
			_uchar4* gpuResult = (_uchar4*)malloc(imageH * imageW * 4);

			if(cpuResult == NULL || gpuResult == NULL)
			{
				fprintf(stderr, "Malloc error\n");
				exit(EXIT_FAILURE);
			}

			cpu_Laplace_withTime(cpuResult, h_Src, imageW, imageH);
			cuda_Laplace(d_dst, cuda_Src, imageW, imageH);

			checkCudaErrors( cudaMemcpy(gpuResult, d_dst, imageW * imageH * 4, cudaMemcpyDeviceToHost) );
			
			if(isArrayEqual(cpuResult, gpuResult, imageH * imageW))
				printf("Results are equal.\n");
			else printf("Results are NOT equal.\n");

			free(cpuResult);
			free(gpuResult);			
		}
		break;

	// save
	case 4:
		{
			_uchar4* buf = (_uchar4*)malloc(imageH * imageW * 4);
			if(buf == NULL)
			{
				fprintf(stderr, "Malloc error\n");
				exit(EXIT_FAILURE);
			}

			checkCudaErrors( cudaMemcpy(buf, d_dst, imageW * imageH * 4, cudaMemcpyDeviceToHost) );

			SaveBMPFile((uchar4*)buf, imageW, imageH, "result.bmp");
			free(buf);
		}
	}

	getLastCudaError("Filtering kernel execution failed.\n");
}


void displayFunc(void)
{
	TColor *d_dst = NULL;
	size_t num_bytes;

	// DEPRECATED: checkCudaErrors(cudaGLMapBufferObject((void**)&d_dst, gl_PBO));
	checkCudaErrors(cudaGraphicsMapResources(1, &cuda_pbo_resource, 0));
	getLastCudaError("cudaGraphicsMapResources failed");
	checkCudaErrors(cudaGraphicsResourceGetMappedPointer((void **)&d_dst, &num_bytes, cuda_pbo_resource));
	getLastCudaError("cudaGraphicsResourceGetMappedPointer failed");

	if(isModeChanged)
		runImageFilters(d_dst, cuda_Src);

	checkCudaErrors(cudaGraphicsUnmapResources(1, &cuda_pbo_resource, 0));

	// Common display code path
	{
		glClear(GL_COLOR_BUFFER_BIT);

		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, imageW, imageH, GL_RGBA, GL_UNSIGNED_BYTE, BUFFER_DATA(0));
		
		glBegin(GL_TRIANGLES);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(2, 0);
		glVertex2f(+3, -1);
		glTexCoord2f(0, 2);
		glVertex2f(-1, +3);
		glEnd();
		glFinish();
	}

	glutSwapBuffers();
}



void shutDown(unsigned char k, int /*x*/, int /*y*/)
{
	switch (k)
	{
	case '\033':
	case 'q':
	case 'Q':
		printf("Shutting down...\n");

		checkCudaErrors( cudaFree(cuda_Src) );

		free(h_Src);
		exit(EXIT_SUCCESS);
		break;

	case '1':
		//printf("Original.\n");
		g_Kernel = 0;
		isModeChanged = true;
		break;

	case '2':
		//printf("Cuda \n");
		g_Kernel = 1;
		isModeChanged = true;
		break;

	case '3':
		//printf("CPU \n");
		g_Kernel = 2;
		isModeChanged = true;
		break;

	case '4':
		//printf("CPU \n");
		g_Kernel = 3;
		isModeChanged = true;
		break;

	case '5':
		//printf("CPU \n");
		g_Kernel = 4;
		isModeChanged = true;
		break;
	}

	displayFunc();
	isModeChanged = false;
}


int initGL(int *argc, char **argv)
{
	printf("Initializing GLUT...\n");
	glutInit(argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	
	glutInitWindowSize(600, 480);
	glutInitWindowPosition(50, 50);
	glutCreateWindow(argv[0]);
	printf("OpenGL window created.\n");

	glewInit();
	//glHint( GL_LINE_SMOOTH_HINT, GL_NICEST ); glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST ); glDisable( GL_LINE_SMOOTH ); glDisable( GL_POLYGON_SMOOTH ); glDisable( GL_MULTISAMPLE );
	printf("Loading extensions: %s\n", glewGetErrorString(glewInit()));

	if (!glewIsSupported("GL_VERSION_1_5 GL_ARB_vertex_buffer_object GL_ARB_pixel_buffer_object"))
	{
		fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
		fprintf(stderr, "This sample requires:\n");
		fprintf(stderr, "  OpenGL version 1.5\n");
		fprintf(stderr, "  GL_ARB_vertex_buffer_object\n");
		fprintf(stderr, "  GL_ARB_pixel_buffer_object\n");
		fflush(stderr);
		return false;
	}

	return 0;
}

void initOpenGLBuffers()
{
	printf("Creating GL texture...\n");
	glEnable(GL_TEXTURE_2D);
	glGenTextures(1, &gl_Tex);
	glBindTexture(GL_TEXTURE_2D, gl_Tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	//glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST )
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, imageW, imageH, 0, GL_RGBA, GL_UNSIGNED_BYTE, h_Src);
	printf("Texture created.\n");


	checkCudaErrors( cudaMalloc((void**)&cuda_Src, imageW * imageH * 4) );
	checkCudaErrors( cudaMemcpy(cuda_Src, h_Src, imageW * imageH * 4, cudaMemcpyHostToDevice) );

	printf("Creating PBO...\n");
	glGenBuffers(1, &gl_PBO);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, gl_PBO);
	glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, imageW * imageH * 4, h_Src, GL_STREAM_COPY);
	//While a PBO is registered to CUDA, it can't be used
	//as the destination for OpenGL drawing calls.
	//But in our particular case OpenGL is only used
	//to display the content of the PBO, specified by CUDA kernels,
	//so we need to register/unregister it only once.
	// DEPRECATED: checkCudaErrors(cudaGLRegisterBufferObject(gl_PBO) );
	checkCudaErrors(cudaGraphicsGLRegisterBuffer(&cuda_pbo_resource, gl_PBO,
		cudaGraphicsMapFlagsWriteDiscard));
	GLenum gl_error = glGetError();

	if (gl_error != GL_NO_ERROR)
	{
#ifdef _WIN32
		char tmpStr[512];
		// NOTE: "%s(%i) : " allows Visual Studio to directly jump to the file at the right line
		// when the user double clicks on the error line in the Output pane. Like any compile error.
		sprintf_s(tmpStr, 255, "\n%s(%i) : GL Error : %s\n\n", __FILE__, __LINE__, gluErrorString(gl_error));
		OutputDebugString(tmpStr);
#endif
		fprintf(stderr, "GL Error in file '%s' in line %d :\n", __FILE__, __LINE__);
		fprintf(stderr, "%s\n", gluErrorString(gl_error));
		exit(EXIT_FAILURE);
	}

	printf("PBO created.\n");
}




int main(int argc, char **argv)
{
	// First load the image, so we know what the size of the image (imageW and imageH)
	printf("Allocating host and CUDA memory and loading image file...\n");
	const char *image_path = argv[1];

	if (image_path == NULL)
	{
		printf("imageDenoisingGL was unable to find and load image file <portrait_noise.bmp>.\nExiting...\n");
		exit(EXIT_FAILURE);
	}

	LoadBMPFile((uchar4**)&h_Src, &imageW, &imageH, image_path);
	printf("Data init done.\n");

	// First initialize OpenGL context, so we can properly set the GL for CUDA.
	// This is necessary in order to achieve optimal performance with OpenGL/CUDA interop.
	initGL(&argc, argv);
	cudaGLSetGLDevice(gpuGetMaxGflopsDeviceId());


	initOpenGLBuffers();

	printf("Starting GLUT main loop...\n");
	printf("Press [1] to view original image\n");
	printf("Press [2] to view image filtered by CUDA\n");
	printf("Press [3] to view image filtered by CPU\n");
	printf("Press [4] to compare CPU and CUDA results\n");
	printf("Press [5] to save picture\n");
	printf("Press [q] to exit\n");

	glutDisplayFunc(displayFunc);
	glutSetWindowTitle("Image filter");
	glutKeyboardFunc(shutDown);

	//displayFunc();
	glutMainLoop();

	cudaDeviceReset();
	exit(EXIT_SUCCESS);
}
